@@ -1,14 +1,15 @@
+use std::borrow::Cow;
 use std::cell::RefCell;
-use std::collections::hash_map::{Entry, HashMap};
+use std::collections::hash_map::HashMap;
+use std::collections::HashSet;
 use std::hash::BuildHasher;
 use std::rc::Rc;
 use std::sync::atomic::AtomicBool;
 use std::sync::Arc;
 
-use ident_case;
-use syn::{self, Lit, Meta, NestedMeta};
+use syn::{Expr, Lit, Meta, NestedMeta};
 
-use {Error, Result};
+use crate::{Error, Result};
 
 /// Create an instance from an item in an attribute declaration.
 ///
@@ -23,6 +24,10 @@ use {Error, Result};
 /// * As a boolean literal, e.g. `foo = true`.
 /// * As a string literal, e.g. `foo = "true"`.
 ///
+/// ## char
+/// * As a char literal, e.g. `foo = '#'`.
+/// * As a string literal consisting of a single character, e.g. `foo = "#"`.
+///
 /// ## String
 /// * As a string literal, e.g. `foo = "hello"`.
 /// * As a raw string literal, e.g. `foo = r#"hello "world""#`.
@@ -96,6 +101,7 @@ pub trait FromMeta: Sized {
         (match *value {
             Lit::Bool(ref b) => Self::from_bool(b.value),
             Lit::Str(ref s) => Self::from_string(&s.value()),
+            Lit::Char(ref ch) => Self::from_char(ch.value()),
             _ => Err(Error::unexpected_lit_type(value)),
         })
         .map_err(|e| e.with_span(value))
@@ -133,6 +139,7 @@ impl FromMeta for bool {
         Ok(true)
     }
 
+    #[allow(clippy::wrong_self_convention)] // false positive
     fn from_bool(value: bool) -> Result<Self> {
         Ok(value)
     }
@@ -150,6 +157,25 @@ impl FromMeta for AtomicBool {
     }
 }
 
+impl FromMeta for char {
+    #[allow(clippy::wrong_self_convention)] // false positive
+    fn from_char(value: char) -> Result<Self> {
+        Ok(value)
+    }
+
+    fn from_string(s: &str) -> Result<Self> {
+        let mut chars = s.chars();
+        let char1 = chars.next();
+        let char2 = chars.next();
+
+        if let (Some(char), None) = (char1, char2) {
+            Ok(char)
+        } else {
+            Err(Error::unexpected_type("string"))
+        }
+    }
+}
+
 impl FromMeta for String {
     fn from_string(s: &str) -> Result<Self> {
         Ok(s.to_string())
@@ -246,6 +272,48 @@ impl<T: syn::parse::Parse, P: syn::parse::Parse> FromMeta for syn::punctuated::P
     }
 }
 
+/// Parsing support for an array, i.e. `example = "[1 + 2, 2 - 2, 3 * 4]"`.
+impl FromMeta for syn::ExprArray {
+    fn from_value(value: &Lit) -> Result<Self> {
+        if let Lit::Str(ref ident) = *value {
+            ident
+                .parse::<syn::ExprArray>()
+                .map_err(|_| Error::unknown_lit_str_value(ident))
+        } else {
+            Err(Error::unexpected_lit_type(value))
+        }
+    }
+}
+
+macro_rules! from_numeric_array {
+    ($ty:ident) => {
+        /// Parsing an unsigned integer array, i.e. `example = "[1, 2, 3, 4]"`.
+        impl FromMeta for Vec<$ty> {
+            fn from_value(value: &Lit) -> Result<Self> {
+                let expr_array = syn::ExprArray::from_value(value)?;
+                // To meet rust <1.36 borrow checker rules on expr_array.elems
+                let v =
+                    expr_array
+                        .elems
+                        .iter()
+                        .map(|expr| match expr {
+                            Expr::Lit(lit) => $ty::from_value(&lit.lit),
+                            _ => Err(Error::custom("Expected array of unsigned integers")
+                                .with_span(expr)),
+                        })
+                        .collect::<Result<Vec<$ty>>>();
+                v
+            }
+        }
+    };
+}
+
+from_numeric_array!(u8);
+from_numeric_array!(u16);
+from_numeric_array!(u32);
+from_numeric_array!(u64);
+from_numeric_array!(usize);
+
 /// Parsing support for paths. This attempts to preserve span information when available,
 /// but also supports parsing strings with the call site as the emitted span.
 impl FromMeta for syn::Path {
@@ -364,43 +432,161 @@ impl<T: FromMeta> FromMeta for RefCell<T> {
     }
 }
 
-impl<V: FromMeta, S: BuildHasher + Default> FromMeta for HashMap<String, V, S> {
-    fn from_list(nested: &[syn::NestedMeta]) -> Result<Self> {
-        let mut map = HashMap::with_capacity_and_hasher(nested.len(), Default::default());
-        for item in nested {
-            if let syn::NestedMeta::Meta(ref inner) = *item {
-                let path = inner.path();
-                let name = path
-                    .segments
-                    .iter()
-                    .map(|s| s.ident.to_string())
-                    .collect::<Vec<String>>()
-                    .join("::");
-                match map.entry(name) {
-                    Entry::Occupied(_) => {
-                        return Err(Error::duplicate_field_path(&path).with_span(inner));
-                    }
-                    Entry::Vacant(entry) => {
-                        // In the error case, extend the error's path, but assume the inner `from_meta`
-                        // set the span, and that subsequently we don't have to.
-                        entry.insert(FromMeta::from_meta(inner).map_err(|e| e.at_path(&path))?);
-                    }
-                }
-            }
-        }
+fn path_to_string(path: &syn::Path) -> String {
+    path.segments
+        .iter()
+        .map(|s| s.ident.to_string())
+        .collect::<Vec<String>>()
+        .join("::")
+}
 
-        Ok(map)
+/// Trait to convert from a path into an owned key for a map.
+trait KeyFromPath: Sized {
+    fn from_path(path: &syn::Path) -> Result<Self>;
+    fn to_display(&self) -> Cow<'_, str>;
+}
+
+impl KeyFromPath for String {
+    fn from_path(path: &syn::Path) -> Result<Self> {
+        Ok(path_to_string(path))
+    }
+
+    fn to_display(&self) -> Cow<'_, str> {
+        Cow::Borrowed(&self)
     }
 }
 
+impl KeyFromPath for syn::Path {
+    fn from_path(path: &syn::Path) -> Result<Self> {
+        Ok(path.clone())
+    }
+
+    fn to_display(&self) -> Cow<'_, str> {
+        Cow::Owned(path_to_string(self))
+    }
+}
+
+impl KeyFromPath for syn::Ident {
+    fn from_path(path: &syn::Path) -> Result<Self> {
+        if path.segments.len() == 1
+            && path.leading_colon.is_none()
+            && path.segments[0].arguments.is_empty()
+        {
+            Ok(path.segments[0].ident.clone())
+        } else {
+            Err(Error::custom("Key must be an identifier").with_span(path))
+        }
+    }
+
+    fn to_display(&self) -> Cow<'_, str> {
+        Cow::Owned(self.to_string())
+    }
+}
+
+macro_rules! hash_map {
+    ($key:ty) => {
+        impl<V: FromMeta, S: BuildHasher + Default> FromMeta for HashMap<$key, V, S> {
+            fn from_list(nested: &[syn::NestedMeta]) -> Result<Self> {
+                // Convert the nested meta items into a sequence of (path, value result) result tuples.
+                // An outer Err means no (key, value) structured could be found, while an Err in the
+                // second position of the tuple means that value was rejected by FromMeta.
+                //
+                // We defer key conversion into $key so that we don't lose span information in the case
+                // of String keys; we'll need it for good duplicate key errors later.
+                let pairs = nested
+                    .iter()
+                    .map(|item| -> Result<(&syn::Path, Result<V>)> {
+                        match *item {
+                            syn::NestedMeta::Meta(ref inner) => {
+                                let path = inner.path();
+                                Ok((
+                                    path,
+                                    FromMeta::from_meta(inner).map_err(|e| e.at_path(&path)),
+                                ))
+                            }
+                            syn::NestedMeta::Lit(_) => Err(Error::unsupported_format("literal")),
+                        }
+                    });
+
+                let mut errors = vec![];
+                // We need to track seen keys separately from the final map, since a seen key with an
+                // Err value won't go into the final map but should trigger a duplicate field error.
+                //
+                // This is a set of $key rather than Path to avoid the possibility that a key type
+                // parses two paths of different values to the same key value.
+                let mut seen_keys = HashSet::with_capacity(nested.len());
+
+                // The map to return in the Ok case. Its size will always be exactly nested.len(),
+                // since otherwise ≥1 field had a problem and the entire map is dropped immediately
+                // when the function returns `Err`.
+                let mut map = HashMap::with_capacity_and_hasher(nested.len(), Default::default());
+
+                for item in pairs {
+                    match item {
+                        Ok((path, value)) => {
+                            let key: $key = match KeyFromPath::from_path(path) {
+                                Ok(k) => k,
+                                Err(e) => {
+                                    errors.push(e);
+
+                                    // Surface value errors even under invalid keys
+                                    if let Err(val_err) = value {
+                                        errors.push(val_err);
+                                    }
+
+                                    continue;
+                                }
+                            };
+
+                            let already_seen = seen_keys.contains(&key);
+
+                            if already_seen {
+                                errors.push(
+                                    Error::duplicate_field(&key.to_display()).with_span(path),
+                                );
+                            }
+
+                            match value {
+                                Ok(_) if already_seen => {}
+                                Ok(val) => {
+                                    map.insert(key.clone(), val);
+                                }
+                                Err(e) => {
+                                    errors.push(e);
+                                }
+                            }
+
+                            seen_keys.insert(key);
+                        }
+                        Err(e) => {
+                            errors.push(e);
+                        }
+                    }
+                }
+
+                if !errors.is_empty() {
+                    return Err(Error::multiple(errors));
+                }
+
+                Ok(map)
+            }
+        }
+    };
+}
+
+// This is done as a macro rather than a blanket impl to avoid breaking backwards compatibility
+// with 0.12.x, while still sharing the same impl.
+hash_map!(String);
+hash_map!(syn::Ident);
+hash_map!(syn::Path);
+
 /// Tests for `FromMeta` implementations. Wherever the word `ignore` appears in test input,
 /// it should not be considered by the parsing.
 #[cfg(test)]
 mod tests {
     use proc_macro2::TokenStream;
-    use syn;
 
-    use {Error, FromMeta, Result};
+    use crate::{Error, FromMeta, Result};
 
     /// parse a string as a syn::Meta instance.
     fn pm(tokens: TokenStream) -> ::std::result::Result<syn::Meta, String> {
@@ -415,7 +601,7 @@ mod tests {
 
     #[test]
     fn unit_succeeds() {
-        assert_eq!(fm::<()>(quote!(ignore)), ());
+        let () = fm::<()>(quote!(ignore));
     }
 
     #[test]
@@ -432,6 +618,15 @@ mod tests {
         assert_eq!(fm::<bool>(quote!(ignore = "false")), false);
     }
 
+    #[test]
+    fn char_succeeds() {
+        // char literal
+        assert_eq!(fm::<char>(quote!(ignore = '😬')), '😬');
+
+        // string literal
+        assert_eq!(fm::<char>(quote!(ignore = "😬")), '😬');
+    }
+
     #[test]
     fn string_succeeds() {
         // cooked form
@@ -442,6 +637,7 @@ mod tests {
     }
 
     #[test]
+    #[allow(clippy::float_cmp)] // we want exact equality
     fn number_succeeds() {
         assert_eq!(fm::<u8>(quote!(ignore = "2")), 2u8);
         assert_eq!(fm::<i16>(quote!(ignore = "-25")), -25i16);
@@ -459,6 +655,7 @@ mod tests {
     }
 
     #[test]
+    #[allow(clippy::float_cmp)] // we want exact equality
     fn float_without_quotes() {
         assert_eq!(fm::<f32>(quote!(ignore = 2.)), 2.0f32);
         assert_eq!(fm::<f32>(quote!(ignore = 2.0)), 2.0f32);
@@ -508,6 +705,68 @@ mod tests {
         assert_eq!(err.to_string(), Error::duplicate_field("hello").to_string());
     }
 
+    #[test]
+    fn hash_map_multiple_errors() {
+        use std::collections::HashMap;
+
+        let err = HashMap::<String, bool>::from_meta(
+            &pm(quote!(ignore(hello, hello = 3, hello = false))).unwrap(),
+        )
+        .expect_err("Duplicates and bad values should error");
+
+        assert_eq!(err.len(), 3);
+        let errors = err.into_iter().collect::<Vec<_>>();
+        assert!(errors[0].has_span());
+        assert!(errors[1].has_span());
+        assert!(errors[2].has_span());
+    }
+
+    #[test]
+    fn hash_map_ident_succeeds() {
+        use std::collections::HashMap;
+        use syn::parse_quote;
+
+        let comparison = {
+            let mut c = HashMap::<syn::Ident, bool>::new();
+            c.insert(parse_quote!(first), true);
+            c.insert(parse_quote!(second), false);
+            c
+        };
+
+        assert_eq!(
+            fm::<HashMap<syn::Ident, bool>>(quote!(ignore(first, second = false))),
+            comparison
+        );
+    }
+
+    #[test]
+    fn hash_map_ident_rejects_non_idents() {
+        use std::collections::HashMap;
+
+        let err: Result<HashMap<syn::Ident, bool>> =
+            FromMeta::from_meta(&pm(quote!(ignore(first, the::second))).unwrap());
+
+        err.unwrap_err();
+    }
+
+    #[test]
+    fn hash_map_path_succeeds() {
+        use std::collections::HashMap;
+        use syn::parse_quote;
+
+        let comparison = {
+            let mut c = HashMap::<syn::Path, bool>::new();
+            c.insert(parse_quote!(first), true);
+            c.insert(parse_quote!(the::second), false);
+            c
+        };
+
+        assert_eq!(
+            fm::<HashMap<syn::Path, bool>>(quote!(ignore(first, the::second = false))),
+            comparison
+        );
+    }
+
     /// Tests that fallible parsing will always produce an outer `Ok` (from `fm`),
     /// and will accurately preserve the inner contents.
     #[test]
@@ -524,4 +783,34 @@ mod tests {
         ));
         fm::<syn::punctuated::Punctuated<syn::Expr, syn::token::Comma>>(quote!(ignore = "a, b, c"));
     }
+
+    #[test]
+    fn test_expr_array() {
+        fm::<syn::ExprArray>(quote!(ignore = "[0x1, 0x2]"));
+        fm::<syn::ExprArray>(quote!(ignore = "[\"Hello World\", \"Test Array\"]"));
+    }
+
+    #[test]
+    fn test_number_array() {
+        assert_eq!(
+            fm::<Vec<u8>>(quote!(ignore = "[16, 0xff]")),
+            vec![0x10, 0xff]
+        );
+        assert_eq!(
+            fm::<Vec<u16>>(quote!(ignore = "[32, 0xffff]")),
+            vec![0x20, 0xffff]
+        );
+        assert_eq!(
+            fm::<Vec<u32>>(quote!(ignore = "[48, 0xffffffff]")),
+            vec![0x30, 0xffffffff]
+        );
+        assert_eq!(
+            fm::<Vec<u64>>(quote!(ignore = "[64, 0xffffffffffffffff]")),
+            vec![0x40, 0xffffffffffffffff]
+        );
+        assert_eq!(
+            fm::<Vec<usize>>(quote!(ignore = "[80, 0xffffffff]")),
+            vec![0x50, 0xffffffff]
+        );
+    }
 }
